from datetime import datetime
import json
import logging
from multiprocessing import Queue, set_start_method
from queue import Empty
import time
import traceback
from uuid import uuid4
from typing import Generator, Iterator

from flask import Blueprint, Response, current_app as app, request, stream_with_context
from pydantic import ValidationError

from workflow_server.core.executor import (
    execute_workflow_pebble_timeout,
    execute_workflow_process_timeout,
    gather_display_context,
    stream_node_pebble_timeout,
    stream_workflow_pebble_timeout,
    stream_workflow_process_timeout,
)
from workflow_server.core.workflow_executor_context import (
    DEFAULT_TIMEOUT_SECONDS,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.utils import (
    VEMBDA_EXECUTION_FULFILLED_EVENT_NAME,
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
    VembdaExecutionInitiatedBody,
    VembdaExecutionInitiatedEvent,
    convert_json_inputs_to_vellum,
    get_version,
)

bp = Blueprint("exec", __name__)

set_start_method("fork", force=True)

logger = logging.getLogger(__name__)


@bp.route("/execute", methods=["POST"])
def exec_route() -> tuple:
    data = request.get_json()
    context = get_workflow_request_context(data)

    app.logger.debug(f"Request received {data.get('execution_id')}")

    try:
        # Pebble uses some sort of managed process pool with timeouts and is like 80ms faster [on live not local]
        # than doing a process yourself. Pebble didn't like it when I used their timeout helper a bunch of times in a
        # big loop and when I used their process pool manually it did other stuff wrong so only used when timeout
        # is the default one
        if context.timeout != DEFAULT_TIMEOUT_SECONDS:
            output = execute_workflow_process_timeout(context)
        else:
            output = execute_workflow_pebble_timeout(context).result()

        resp = {
            "log": output["log"],
            "result": output["result"],
            "stderr": output["stderr"],
            "exit_code": output["exit_code"],
            "timed_out": False,
        }

        app.logger.debug(f"Request complete {data.get('execution_id')}")
        return resp, 200

    except TimeoutError:
        resp = {
            "log": "",
            "result": "",
            "stderr": "",
            "exit_code": -1,
            "timed_out": True,
        }
        return resp, 200
    except Exception as e:
        app.logger.exception(e)

        return {
            "result": "",
            "stderr": traceback.format_exc(),
            "exit_code": -1,
            "timed_out": False,
        }, 200


@bp.route("/stream", methods=["POST"])
def stream_workflow_route() -> Response:
    data = request.get_json()

    try:
        context = get_workflow_request_context(data)
    except ValidationError as e:
        error_message = e.errors()[0]["msg"]
        error_location = e.errors()[0]["loc"]

        return Response(
            json.dumps({"detail": f"Invalid context: {error_message} at {error_location}"}),
            status=400,
            content_type="application/json",
        )

    app.logger.debug(f"Stream received {data.get('execution_id')}")
    try:
        display_context = gather_display_context(context)
    except Exception:
        logger.exception("Unable to Parse Workflow Display Context")
        display_context = None

    # Create this event up here so timestamps are fully from the start to account for any unknown overhead
    vembda_initiated_event = VembdaExecutionInitiatedEvent(
        id=uuid4(),
        timestamp=datetime.now(),
        trace_id=context.trace_id,
        span_id=context.execution_id,
        body=VembdaExecutionInitiatedBody(
            sdk_version=get_version().get("sdk_version"),
            server_version=get_version().get("server_version"),
            display_context=display_context,
        ),
        parent=None,
    )

    if context.timeout != DEFAULT_TIMEOUT_SECONDS:
        process_queue: Queue[dict] = Queue()
        stream_workflow_process_timeout(
            executor_context=context,
            queue=process_queue,
        )

        def process_events() -> Iterator[dict]:
            while True:
                try:
                    event = process_queue.get(timeout=context.timeout)
                    yield event

                    if event.get("name") == VEMBDA_EXECUTION_FULFILLED_EVENT_NAME:
                        break
                except Empty:
                    continue
                except Exception as e:
                    # This happens when theres a problem with process execution itself not the workflow runner
                    vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                        id=uuid4(),
                        timestamp=datetime.now(),
                        trace_id=context.trace_id,
                        span_id=context.execution_id,
                        body=VembdaExecutionFulfilledBody(
                            exit_code=-1,
                            stderr="Internal Server Error",
                            container_overhead_latency=context.container_overhead_latency,
                        ),
                        parent=None,
                    )
                    yield vembda_fulfilled_event.model_dump(mode="json")
                    app.logger.exception(e)
                    break

        workflow_events = process_events()
    else:
        pebble_queue: Queue[dict] = Queue()
        stream_future = stream_workflow_pebble_timeout(
            executor_context=context,
            queue=pebble_queue,
        )

        def pebble_events() -> Iterator[dict]:
            while True:
                try:
                    event = pebble_queue.get(timeout=context.timeout)
                except Empty:
                    if stream_future.exception() is not None:
                        # This happens when theres a problem with the stream function call
                        # itself not the workflow runner
                        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                            id=uuid4(),
                            timestamp=datetime.now(),
                            trace_id=context.trace_id,
                            span_id=context.execution_id,
                            body=VembdaExecutionFulfilledBody(
                                exit_code=-1,
                                stderr="Internal Server Error",
                                container_overhead_latency=context.container_overhead_latency,
                            ),
                            parent=None,
                        )
                        yield vembda_fulfilled_event.model_dump(mode="json")
                        app.logger.exception(stream_future.exception())
                        break
                    else:
                        continue

                yield event
                if event.get("name") == VEMBDA_EXECUTION_FULFILLED_EVENT_NAME:
                    break

        workflow_events = pebble_events()

    def generator() -> Generator[str, None, None]:
        yield json.dumps(vembda_initiated_event.model_dump(mode="json"))

        for index, row in enumerate(workflow_events):
            yield "\n"
            yield json.dumps(row)

    resp = Response(stream_with_context(generator()), status=200, content_type='application/x-ndjson"')
    return resp


@bp.route("/stream-node", methods=["POST"])
def stream_node_route() -> Response:
    data = request.get_json()

    try:
        context = get_node_request_context(data)
    except ValidationError as e:
        error_message = e.errors()[0]["msg"]
        error_location = e.errors()[0]["loc"]
        return Response(
            json.dumps({"detail": f"Invalid context: {error_message} at {error_location}"}),
            status=400,
            content_type="application/json",
        )

    # Create this event up here so timestamps are fully from the start to account for any unknown overhead
    vembda_initiated_event = VembdaExecutionInitiatedEvent(
        id=uuid4(),
        timestamp=datetime.now(),
        trace_id=context.trace_id,
        span_id=context.execution_id,
        body=VembdaExecutionInitiatedBody(
            sdk_version=get_version().get("sdk_version"),
            server_version=get_version().get("server_version"),
        ),
        parent=None,
    )

    app.logger.debug(f"Node stream received {data.get('execution_id')}")

    pebble_queue: Queue[dict] = Queue()
    stream_future = stream_node_pebble_timeout(
        executor_context=context,
        queue=pebble_queue,
    )

    def node_events() -> Iterator[dict]:
        while True:
            try:
                event = pebble_queue.get(timeout=context.timeout)
            except Empty:
                if stream_future.exception() is not None:
                    # This happens when theres a problem with the stream function call
                    # itself not the workflow runner
                    vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                        id=uuid4(),
                        timestamp=datetime.now(),
                        trace_id=context.trace_id,
                        span_id=context.execution_id,
                        body=VembdaExecutionFulfilledBody(
                            exit_code=-1,
                            stderr="Internal Server Error",
                            container_overhead_latency=context.container_overhead_latency,
                        ),
                        parent=None,
                    )
                    yield vembda_fulfilled_event.model_dump(mode="json")
                    app.logger.exception(stream_future.exception())
                    break
                else:
                    continue

            yield event
            if event.get("name") == VEMBDA_EXECUTION_FULFILLED_EVENT_NAME:
                break

    def generator() -> Generator[str, None, None]:
        yield json.dumps(vembda_initiated_event.model_dump(mode="json"))

        for index, row in enumerate(node_events()):
            yield "\n"
            yield json.dumps(row)

    resp = Response(stream_with_context(generator()), status=200, content_type='application/x-ndjson"')
    return resp


@bp.route("/version", methods=["GET"])
def get_version_route() -> tuple:
    resp = get_version()

    return resp, 200


def get_workflow_request_context(data) -> WorkflowExecutorContext:
    # not sure if this is the filter we want to pass forward?
    context_data = {
        **data,
        "inputs": convert_json_inputs_to_vellum(data["inputs"]),
        "trace_id": uuid4(),
        "request_start_time": time.time_ns(),
    }

    return WorkflowExecutorContext.model_validate(context_data)


def get_node_request_context(data) -> NodeExecutorContext:
    context_data = {
        **data,
        "inputs": convert_json_inputs_to_vellum(data["inputs"]),
        "trace_id": uuid4(),
        "request_start_time": time.time_ns(),
    }

    return NodeExecutorContext.model_validate(context_data)



import re

from workflow_server.server import create_app


def test_version_route():
    flask_app = create_app()

    with flask_app.test_client() as test_client:
        response = test_client.get("/workflow/version")
        assert response.status_code == 200
        assert re.match(r"[0-9]*\.[0-9]*\.[0-9]*", response.json["sdk_version"])
        assert response.json["server_version"] == "local"



import os

import gunicorn.app.base

from workflow_server.server import app


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items() if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def start():
    # gevent didn't play nice with pebble and processes so just gthread here
    options = {
        "bind": "0.0.0.0:8000",
        "workers": int(os.getenv("GUNICORN_WORKERS", 2)),
        "threads": int(os.getenv("GUNICORN_THREADS", 8)),
        "worker_class": "gthread",
        "timeout": int(os.getenv("GUNICORN_TIMEOUT", 1800)),
    }
    StandaloneApplication(app, options).run()



from importlib.metadata import version
import json
import re
from typing import List, Literal, Optional

from vellum_ee.workflows.display.types import WorkflowEventDisplayContext

from vellum import (
    ChatMessageRequest,
    FunctionCallRequest,
    SearchResultRequest,
    VellumErrorRequest,
    VellumImageRequest,
    VellumValueRequest,
)
from vellum.client.core import UniversalBaseModel
from vellum.workflows import BaseWorkflow
from vellum.workflows.events.types import BaseEvent
from vellum.workflows.nodes import BaseNode
from workflow_server.config import is_development

VEMBDA_EXECUTION_INITIATED_EVENT_NAME = "vembda.execution.initiated"
VEMBDA_EXECUTION_FULFILLED_EVENT_NAME = "vembda.execution.fulfilled"


class VembdaExecutionFulfilledBody(UniversalBaseModel):
    exit_code: int = 0
    log: str = ""
    stderr: str = ""
    timed_out: bool = False
    container_overhead_latency: float = 0


class VembdaExecutionFulfilledEvent(BaseEvent):
    name: Literal["vembda.execution.fulfilled"] = VEMBDA_EXECUTION_FULFILLED_EVENT_NAME  # type: ignore
    body: VembdaExecutionFulfilledBody


class VembdaExecutionInitiatedBody(UniversalBaseModel):
    sdk_version: str
    server_version: str
    display_context: Optional[WorkflowEventDisplayContext] = None


class VembdaExecutionInitiatedEvent(BaseEvent):
    name: Literal["vembda.execution.initiated"] = VEMBDA_EXECUTION_INITIATED_EVENT_NAME  # type: ignore
    body: VembdaExecutionInitiatedBody


VembdaExecutionFulfilledEvent.model_rebuild(
    # Not sure why this is needed, but it is required for the VembdaExecutionFulfilledEvent to be
    # properly rebuilt with the recursive types.
    _types_namespace={
        "BaseWorkflow": BaseWorkflow,
        "BaseNode": BaseNode,
    },
)


def convert_json_inputs_to_vellum(inputs: List[dict]):
    vellum_inputs = {}

    for input in inputs:
        value = input["value"]
        name = to_python_safe_snake_case(input["name"])
        type = input["type"]

        if type == "CHAT_HISTORY":
            vellum_inputs[name] = [ChatMessageRequest.model_validate_json(json.dumps(item)) for item in value]
        elif type == "FUNCTION_CALL":
            vellum_inputs[name] = FunctionCallRequest.model_validate_json(json.dumps(value))
        elif type == "IMAGE":
            vellum_inputs[name] = VellumImageRequest.model_validate_json(json.dumps(value))
        elif type == "SEARCH_RESULTS":
            vellum_inputs[name] = SearchResultRequest.model_validate_json(json.dumps(value))
        elif type == "ERROR":
            vellum_inputs[name] = VellumErrorRequest.model_validate_json(json.dumps(value))
        elif type == "ARRAY":
            vellum_inputs[name] = [VellumValueRequest.model_validate_json(json.dumps(item)) for item in value]
        else:
            vellum_inputs[name] = value

    return vellum_inputs


def get_version():
    return {
        "sdk_version": version("vellum-ai"),
        "server_version": "local" if is_development() else version("vellum-workflow-server"),
    }


def to_python_safe_snake_case(string: str, safety_prefix: str = "_") -> str:
    # Strip special characters from start of string
    cleaned_str = re.sub(r"^[^a-zA-Z0-9_]+", "", string)

    # Check if cleaned string starts with a number
    starts_with_unsafe = bool(re.match(r"^\d", cleaned_str))

    # Convert to snake case
    snake_case = re.sub(r"([a-z])([A-Z])", r"\1_\2", cleaned_str)  # Insert underscore between lower and upper case
    snake_case = re.sub(r"[^a-zA-Z0-9]+", "_", snake_case)  # Replace any non-alphanumeric chars with underscore
    snake_case = re.sub(r"^_+|_+$", "", snake_case)  # Remove leading/trailing underscores
    snake_case = snake_case.lower()

    # Add safety prefix if needed
    cleaned_safety_prefix = (
        "_" if safety_prefix == "_" else f"{safety_prefix}{'_' if not safety_prefix.endswith('_') else ''}"
    )
    return f"{cleaned_safety_prefix}{snake_case}" if starts_with_unsafe else snake_case



from vellum import ChatMessageRequest
from workflow_server.utils.utils import convert_json_inputs_to_vellum


def test_map_error_message__no_remap():
    inputs = [
        {"type": "STRING", "name": "test", "value": "<example-string-value>"},
        {"type": "NUMBER", "name": "test2", "value": 5},
        {"type": "JSON", "name": "test3", "value": {"example-key": "example-value"}},
        {"type": "CHAT_HISTORY", "name": "chat_history", "value": [{"role": "USER", "text": "<example-user-text>"}]},
    ]

    expected = {
        "chat_history": [
            ChatMessageRequest(
                text="<example-user-text>",
                role="USER",
                content=None,
                source=None,
            )
        ],
        "test": "<example-string-value>",
        "test2": 5,
        "test3": {"example-key": "example-value"},
    }

    actual = convert_json_inputs_to_vellum(inputs)

    assert expected == actual


def test_input_variables_with_uppercase_gets_sanitized():
    inputs = [
        {"type": "STRING", "name": "Foo", "value": "<example-string-value>"},
        {"type": "STRING", "name": "Foo-Var", "value": "<another-example-string-value>"},
    ]

    expected = {
        "foo": "<example-string-value>",
        "foo_var": "<another-example-string-value>",
    }

    actual = convert_json_inputs_to_vellum(inputs)

    assert expected == actual



import json

from flask import Request, Response
import jwt
from jwt import ExpiredSignatureError

from workflow_server.config import VEMBDA_PUBLIC_KEY, is_development


class AuthMiddleware:
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        try:
            request = Request(environ)
            if not request.path.startswith("/healthz") and not is_development():
                token = request.headers.get("X-Vembda-Signature")
                jwt.decode(token, VEMBDA_PUBLIC_KEY, algorithms=["RS256"])

        except ExpiredSignatureError:
            res = Response(
                json.dumps({"detail": "Signature token has expired. Please obtain a new token."}),
                mimetype="application/json",
                status=401,
            )
            return res(environ, start_response)
        except Exception as e:
            res = Response(
                json.dumps({"detail": f"Invalid signature token {str(e)}"}), mimetype="application/json", status=401
            )
            return res(environ, start_response)

        return self.app(environ, start_response)



from importlib.metadata import version
import json
from unittest import mock
from uuid import uuid4

from workflow_server.server import create_app


def test_stream_workflow_route__happy_path():
    # GIVEN a flask app
    flask_app = create_app()

    # AND a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [],
        "workspace_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow

class Workflow(BaseWorkflow):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
        },
    }

    # WHEN we call the stream route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code
        events = [json.loads(line) for line in response.data.decode().split("\n") if line]

    # THEN we get a 200 response
    assert status_code == 200, events

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
            "display_context": None,
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__happy_path_with_inputs():
    # GIVEN a flask app
    flask_app = create_app()

    # AND a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [
            {"name": "foo", "type": "STRING", "value": "hello"},
        ],
        "workspace_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

class Inputs(BaseInputs):
    foo: str
""",
        },
    }

    # WHEN we call the stream route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

    # THEN we get a 200 response
    assert status_code == 200, response.json()
    events = [json.loads(line) for line in response.data.decode().split("\n") if line]

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
            "display_context": None,
        },
    }

    assert events[1]["name"] == "workflow.execution.initiated", events[1]
    assert events[2]["name"] == "workflow.execution.fulfilled", events[2]

    assert events[3] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[3]["body"] == {
        "exit_code": 0,
        "log": "",
        "stderr": "",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 4


def test_stream_workflow_route__bad_indent_in_inputs_file():
    # GIVEN a flask app
    flask_app = create_app()

    # AND a valid request body
    span_id = uuid4()
    request_body = {
        "timeout": 360,
        "execution_id": str(span_id),
        "inputs": [
            {"name": "foo", "type": "STRING", "value": "hello"},
        ],
        "workspace_api_key": "test",
        "module": "workflow",
        "files": {
            "__init__.py": "",
            "workflow.py": """\
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState
from .inputs import Inputs

class Workflow(BaseWorkflow[Inputs, BaseState]):
    class Outputs(BaseWorkflow.Outputs):
        foo = "hello"
""",
            "inputs.py": """\
from vellum.workflows.inputs import BaseInputs

  class Inputs(BaseInputs):
     foo: str
""",
        },
    }

    # WHEN we call the stream route
    with flask_app.test_client() as test_client:
        response = test_client.post("/workflow/stream", json=request_body)
        status_code = response.status_code

    # THEN we get a 200 response
    assert status_code == 200, response.json()
    events = [json.loads(line) for line in response.data.decode().split("\n") if line]

    # THEN we get the expected events
    assert events[0] == {
        "id": mock.ANY,
        "trace_id": mock.ANY,
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.initiated",
        "body": {
            "sdk_version": version("vellum-ai"),
            "server_version": "local",
            "display_context": None,
        },
    }

    assert events[1] == {
        "id": mock.ANY,
        "trace_id": events[0]["trace_id"],
        "span_id": str(span_id),
        "timestamp": mock.ANY,
        "api_version": "2024-10-25",
        "parent": None,
        "name": "vembda.execution.fulfilled",
        "body": mock.ANY,
    }
    assert events[1]["body"] == {
        "exit_code": -1,
        "log": "",
        "stderr": "Failed to initialize workflow inputs: unexpected indent (<string>, line 3)",
        "timed_out": False,
        "container_overhead_latency": mock.ANY,
    }

    assert len(events) == 2



from datetime import datetime
import json
import os
from uuid import uuid4

from workflow_server.api.workflow_view import get_workflow_request_context
from workflow_server.core.executor import stream_workflow
from workflow_server.utils.utils import (
    VembdaExecutionFulfilledBody,
    VembdaExecutionFulfilledEvent,
    VembdaExecutionInitiatedBody,
    VembdaExecutionInitiatedEvent,
    get_version,
)

_EVENT_LINE = "--event--"


def run():
    context = None

    try:
        input_raw = ""
        while "--vellum-input-stop--" not in input_raw:
            # os/read they come in chunks of idk length, not as lines
            input_raw += os.read(0, 100_000_000).decode("utf-8")

        split_input = input_raw.split("\n--vellum-input-stop--\n")
        input_json = split_input[0]

        input_data = json.loads(input_json)
        context = get_workflow_request_context(input_data)

        print("--vellum-output-start--")  # noqa: T201

        vembda_initiated_event = VembdaExecutionInitiatedEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=context.trace_id,
            span_id=context.execution_id,
            body=VembdaExecutionInitiatedBody(
                sdk_version=get_version().get("sdk_version"),
                server_version=get_version().get("server_version"),
            ),
            parent=None,
        )
        vembda_initiated_event.model_dump(mode="json")

        print(f"{_EVENT_LINE}{json.dumps(vembda_initiated_event)}")  # noqa: T201

        for line in stream_workflow(context, output={}, disable_redirect=True):
            print(f"{_EVENT_LINE}{json.dumps(line)}")  # noqa: T201
    except Exception:
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=context.trace_id if context else None,
            span_id=context.execution_id if context else None,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                stderr="Internal Server Error",
            ),
            parent=None,
        )
        event = json.dumps(vembda_fulfilled_event.model_dump(mode="json"))
        print(f"{_EVENT_LINE}{event}")  # noqa: T201



from contextlib import redirect_stdout
from datetime import datetime
from functools import lru_cache
import importlib
from io import StringIO
import json
import logging
from multiprocessing import Manager, Process, Queue
import os
import random
import string
import sys
import time
from traceback import format_exc
from uuid import uuid4
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator, Optional, Tuple

from pebble import concurrent
from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder

from vellum import Vellum, VellumEnvironment
from vellum.workflows import BaseWorkflow
from vellum.workflows.events.types import BaseEvent
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode
from vellum.workflows.state.base import BaseState, StateMeta
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.workflows.event_filters import all_workflow_event_filter
from workflow_server.core.workflow_executor_context import (
    DEFAULT_TIMEOUT_SECONDS,
    BaseExecutorContext,
    NodeExecutorContext,
    WorkflowExecutorContext,
)
from workflow_server.utils.exceptions import WorkflowInitializationException
from workflow_server.utils.log_proxy import redirect_log
from workflow_server.utils.utils import VembdaExecutionFulfilledBody, VembdaExecutionFulfilledEvent

if TYPE_CHECKING:
    from vellum_ee.workflows.display.types import WorkflowEventDisplayContext

logger = logging.getLogger(__name__)


@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
def execute_workflow_pebble_timeout(executor_context: WorkflowExecutorContext) -> dict:
    return execute_workflow(executor_context, output={})


def execute_workflow_process_timeout(executor_context: WorkflowExecutorContext) -> dict:
    try:
        with Manager() as manager:
            output = manager.dict()

            p = Process(
                target=execute_workflow,
                args=(
                    executor_context,
                    output,
                ),
            )
            p.start()
            p.join(executor_context.timeout)
            if p.is_alive():
                p.kill()

            if output.get("exit_code") is None:
                return {
                    "log": "",
                    "result": {},
                    "stderr": "",
                    "exit_code": -1,
                    "timed_out": True,
                }

            return {
                "log": output["log"],
                "result": output["result"],
                "stderr": output["stderr"],
                "exit_code": output["exit_code"],
                "timed_out": False,
            }

    except Exception as e:
        logger.exception(e)

        return {
            "result": {},
            "stderr": format_exc(),
            "exit_code": -1,
            "timed_out": False,
        }


def execute_workflow(executor_context: WorkflowExecutorContext, output: dict) -> dict:
    log_redirect = StringIO()

    try:
        # We need to namespace to avoid any collisions with the pebble managed process pool that may or may not
        # reuse processes and also this avoids collisions with local packages and also fixes the virtual loader
        # not working with relative import paths. We also do this renaming on the fly in the virtual file finder
        # to avoid copying the files dict and costing extra latency and ram.
        namespace = _get_file_namespace(executor_context)
        sys.meta_path.append(VirtualFileFinder(executor_context.files, namespace))
        inputs = importlib.import_module(f"{namespace}.inputs")

        with redirect_stdout(log_redirect):
            result = _create_workflow(
                executor_context=executor_context,
                namespace=namespace,
            ).run(
                inputs=inputs.Inputs(**executor_context.inputs),
            )

        log = log_redirect.getvalue()

        exec_output = result
        output["result"] = exec_output.model_dump(mode="json")
        output["log"] = log
        output["stderr"] = ""
        output["exit_code"] = 0
    except Exception:
        output["result"] = ""
        output["exit_code"] = -1
        output["log"] = log_redirect.getvalue()
        output["stderr"] = format_exc()

    return output


@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
# type ignore since pebble annotation changes return type
def stream_workflow_pebble_timeout(
    executor_context: WorkflowExecutorContext,
    queue: Queue,
) -> None:
    _stream_workflow_wrapper(
        executor_context=executor_context,
        output={},
        queue=queue,
    )


@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
# type ignore since pebble annotation changes return type
def stream_node_pebble_timeout(
    executor_context: NodeExecutorContext,
    queue: Queue,
) -> None:
    _stream_node_wrapper(
        executor_context=executor_context,
        output={},
        queue=queue,
    )


def _stream_node_wrapper(executor_context: NodeExecutorContext, output: dict, queue: Queue) -> None:
    try:
        for event in stream_node(executor_context=executor_context, output=output):
            queue.put(event)
    except Exception as e:
        logger.exception(e)
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr="Internal Server Error",
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )


def _stream_workflow_wrapper(executor_context: WorkflowExecutorContext, output: dict, queue: Queue) -> None:
    try:
        for event in stream_workflow(executor_context=executor_context, output=output):
            queue.put(event)
    except WorkflowInitializationException as e:
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr=str(e),
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )
    except Exception as e:
        logger.exception(e)
        queue.put(
            VembdaExecutionFulfilledEvent(
                id=uuid4(),
                timestamp=datetime.now(),
                trace_id=executor_context.trace_id,
                span_id=executor_context.execution_id,
                body=VembdaExecutionFulfilledBody(
                    exit_code=-1,
                    stderr="Internal Server Error",
                    container_overhead_latency=executor_context.container_overhead_latency,
                ),
                parent=None,
            ).model_dump(mode="json")
        )


def stream_workflow_process_timeout(executor_context: WorkflowExecutorContext, queue: Queue) -> None:
    try:
        with Manager() as manager:
            output = manager.dict()

            p = Process(
                target=_stream_workflow_wrapper,
                args=(
                    executor_context,
                    output,
                    queue,
                ),
            )
            p.start()
            p.join(executor_context.timeout)
            if p.is_alive():
                p.kill()

            if output.get("exit_code") is None:
                vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
                    id=uuid4(),
                    timestamp=datetime.now(),
                    trace_id=executor_context.trace_id,
                    span_id=executor_context.execution_id,
                    body=VembdaExecutionFulfilledBody(
                        exit_code=-1,
                        timed_out=True,
                        container_overhead_latency=executor_context.container_overhead_latency,
                    ),
                    parent=None,
                )
                queue.put(vembda_fulfilled_event.model_dump(mode="json"))

    except Exception as e:
        logger.exception(e)

        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                stderr=format_exc(),
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        queue.put(vembda_fulfilled_event.model_dump(mode="json"))


def stream_workflow(
    executor_context: WorkflowExecutorContext,
    output: dict,
    disable_redirect: bool = False,
) -> Iterator[dict]:
    workspace, namespace = gather_workflow(executor_context)
    workflow_inputs = _get_workflow_inputs(executor_context)

    def call_workflow() -> Generator[dict[str, Any], Any, None]:
        stream = _create_workflow(
            executor_context=executor_context,
            namespace=namespace,
        ).stream(
            inputs=workflow_inputs,
            event_filter=all_workflow_event_filter,
        )

        first = True
        for event in stream:
            if first:
                executor_context.stream_start_time = time.time_ns()
                first = False

            yield _dump_event(
                event=event,
                executor_context=executor_context,
            )

    return _call_stream(
        executor_context=executor_context,
        output=output,
        stream_generator=call_workflow,
        disable_redirect=disable_redirect,
    )


def stream_node(
    executor_context: NodeExecutorContext,
    output: dict,
    disable_redirect: bool = False,
) -> Iterator[dict]:
    namespace = _get_file_namespace(executor_context)

    def call_node() -> Generator[dict[str, Any], Any, None]:
        sys.meta_path.append(VirtualFileFinder(executor_context.files, namespace))
        workflow_context = _create_workflow_context(executor_context)
        node_module = importlib.import_module(f"{namespace}.{executor_context.node_module}")

        Node = getattr(node_module, executor_context.node_name)

        workflow_inputs = _get_workflow_inputs(executor_context)

        State = (
            importlib.import_module(f"{namespace}.state").State if executor_context.files.get("state.py") else BaseState
        )

        if executor_context.state:
            state = State.parse_raw(executor_context.state)
        else:
            state = State(
                meta=StateMeta(workflow_inputs=workflow_inputs),
            )

        node = Node(
            state=state,
            context=workflow_context,
        )

        executor_context.stream_start_time = time.time_ns()
        node_outputs = node.run()

        if isinstance(node_outputs, (Iterator)):
            for node_output in node_outputs:
                yield json.loads(json.dumps(node_output, default=vars))
        else:
            yield json.loads(json.dumps(node_outputs, default=vars))

    return _call_stream(
        executor_context=executor_context,
        output=output,
        stream_generator=call_node,
        disable_redirect=disable_redirect,
    )


def _call_stream(
    executor_context: BaseExecutorContext,
    output: dict,
    stream_generator: Callable[[], Generator[dict[str, Any], Any, None]],
    disable_redirect: bool = False,
) -> Iterator[dict]:
    log_redirect: Optional[StringIO] = None

    if not disable_redirect:
        log_redirect = redirect_log()

    try:
        yield from stream_generator()

        output["exit_code"] = 0
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=0,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr="",
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")

    except Exception:
        output["exit_code"] = -1
        vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
            id=uuid4(),
            timestamp=datetime.now(),
            trace_id=executor_context.trace_id,
            span_id=executor_context.execution_id,
            body=VembdaExecutionFulfilledBody(
                exit_code=-1,
                log=log_redirect.getvalue() if log_redirect else "",
                stderr=format_exc(),
                container_overhead_latency=executor_context.container_overhead_latency,
            ),
            parent=None,
        )
        yield vembda_fulfilled_event.model_dump(mode="json")


def _create_workflow(executor_context, namespace) -> BaseWorkflow:
    workflow_context = _create_workflow_context(executor_context)
    Workflow = BaseWorkflow.load_from_module(namespace)
    VembdaExecutionFulfilledEvent.model_rebuild(
        # Not sure why this is needed, but it is required for the VembdaExecutionFulfilledEvent to be
        # properly rebuilt with the recursive types.
        _types_namespace={
            "BaseWorkflow": BaseWorkflow,
            "BaseNode": BaseNode,
        },
    )

    return Workflow(context=workflow_context)


def _create_workflow_context(executor_context) -> WorkflowContext:
    if os.getenv("USE_LOCAL_VELLUM_API") == "true":
        VELLUM_API_URL = "http://localhost:8000"
        environment = VellumEnvironment(
            default=VELLUM_API_URL,
            documents=VELLUM_API_URL,
            predict=VELLUM_API_URL,
        )
    else:
        environment = VellumEnvironment.PRODUCTION

    return WorkflowContext(
        vellum_client=Vellum(
            api_key=executor_context.workspace_api_key,
            environment=environment,
        ),
        parent_context=executor_context.parent_context,
    )


def _get_file_namespace(executor_context: BaseExecutorContext) -> str:
    return str(executor_context.execution_id) or "".join(
        random.choice(string.ascii_letters + string.digits) for i in range(14)
    )


def _dump_event(event: BaseEvent, executor_context: BaseExecutorContext) -> dict:
    module_base = executor_context.module.split(".")
    dump = event.model_dump(mode="json")
    if dump["name"] in {
        "workflow.execution.initiated",
        "workflow.execution.fulfilled",
        "workflow.execution.rejected",
        "workflow.execution.streaming",
        "workflow.execution.paused",
        "workflow.execution.resumed",
    }:
        dump["body"]["workflow_definition"]["module"] = module_base + dump["body"]["workflow_definition"]["module"][1:]
    elif dump["name"] in {
        "node.execution.initiated",
        "node.execution.fulfilled",
        "node.execution.rejected",
        "node.execution.streaming",
        "node.execution.paused",
        "node.execution.resumed",
    }:
        dump["body"]["node_definition"]["module"] = module_base + dump["body"]["node_definition"]["module"][1:]

    return dump


def _get_workflow_inputs(executor_context: WorkflowExecutorContext) -> Optional[BaseInputs]:
    if not executor_context.inputs:
        return None

    if not executor_context.files.get("inputs.py"):
        return None

    namespace = _get_file_namespace(executor_context)
    inputs_module_path = f"{namespace}.inputs"
    try:
        inputs_module = importlib.import_module(inputs_module_path)
    except Exception as e:
        raise WorkflowInitializationException(f"Failed to initialize workflow inputs: {e}") from e

    if not hasattr(inputs_module, "Inputs"):
        raise WorkflowInitializationException(
            f"Inputs module {inputs_module_path} does not have a required Inputs class"
        )

    if not issubclass(inputs_module.Inputs, BaseInputs):
        raise WorkflowInitializationException(
            f"""The class {inputs_module_path}.Inputs was expected to be a subclass of BaseInputs, \
but found {inputs_module.Inputs.__class__.__name__}"""
        )

    return inputs_module.Inputs(**executor_context.inputs)


@lru_cache(maxsize=2)
def gather_workflow(context: WorkflowExecutorContext) -> Tuple[BaseWorkflow, str]:
    try:
        namespace = _get_file_namespace(context)
        sys.meta_path.append(VirtualFileFinder(context.files, namespace))
        workflow = _create_workflow(
            executor_context=context,
            namespace=namespace,
        )
        return workflow, namespace
    except Exception as e:
        raise WorkflowInitializationException(f"Failed to initialize workflow inputs: {e}") from e


def gather_display_context(context: WorkflowExecutorContext) -> "WorkflowEventDisplayContext":
    workflow, namespace = gather_workflow(context)
    display_context = BaseWorkflowDisplay.gather_event_display_context(namespace, workflow.__class__)
    return display_context



from uuid import UUID
from typing import Optional

from vellum.client.core import UniversalBaseModel
from vellum.workflows.events.types import ParentContext

DEFAULT_TIMEOUT_SECONDS = 60 * 30


class BaseExecutorContext(UniversalBaseModel):
    inputs: dict
    state: Optional[dict] = None
    timeout: int = DEFAULT_TIMEOUT_SECONDS
    files: dict[str, str]
    workspace_api_key: str
    execution_id: UUID
    trace_id: UUID
    module: str
    parent_context: Optional[ParentContext] = None
    request_start_time: int
    stream_start_time: int = 0

    @property
    def container_overhead_latency(self) -> int:
        return self.stream_start_time - self.request_start_time if self.stream_start_time else -1

    def __hash__(self):
        # do we think we need anything else for a unique hash for caching?
        return hash(str(self.execution_id))


class WorkflowExecutorContext(BaseExecutorContext):
    pass


class NodeExecutorContext(BaseExecutorContext):
    node_module: str
    node_name: str
